#include "stdafx.h"

#include "P3PMatchInfo.h"
#include "PolicyHandler.h"
#include "ReadHTTPFile.h"
#include "Globals.h"
#include "P3PPReferences.h"
#include "resource.h"
#include "PatMatch.h"


void P3PMatchInfo::CheckValid()
{
	if( !valid_object )
	{
		CleanInfo();
		THROW("STOP");
	}
}

void P3PMatchInfo::AddEmbededURL(const GURL &url, const GString &type)
{
	CheckValid();

	GCriticalSectionLock lock(&InfoLock);
	if( GPosition pos = embededURLs.contains(url) )
	{
		return;
	}
	else
	{
		embededURLs[url] = URLTypeInfo(type);
	}
}

void P3PMatchInfo::CheckEmbeds()
{
	CheckValid();

	GCriticalSectionLock lock(&InfoLock);
	GMap<GURL, bool> prefurls;

	TRY
	{
	for( GPosition pos=embededURLs; pos; ++pos)
	{
		CheckValid();
		prefurls[embededURLs.key(pos).host()+"/w3c/p3p.xml"] = false;
	}

	for( pos = prefurls; pos; ++pos )
	{
		CheckValid();
		TRY
		{
			g_policy_handler.HandlePolicyReferenceURL(prefurls.key(pos), prefurls.key(pos));
		}
		CATCH(ex)
		{
		}
		ENDCATCH;
	}

	for( pos=embededURLs; pos; ++pos)
	{
		CheckValid();
		const GURL &embedurl = embededURLs.key(pos);
		GURL policyurl = g_policy_handler.FindPolicy(embedurl, "GET");
		if( policyurl.is_empty() )
		{	///check for applicable policy using hints
			GList<GURL> hint_list;
			g_policy_handler.GetApplicableHints(embedurl, hint_list);
			for(GPosition hintpos = hint_list; hintpos && policyurl.is_empty(); ++hintpos)
			{
				g_policy_handler.HandlePolicyReferenceURL(hint_list[hintpos], hint_list[hintpos]);
				policyurl = g_policy_handler.FindPolicy(embedurl, "GET");
			}
		}
		embededURLs[pos].URL = policyurl;
		if( !policyurl.is_empty() )
			policyURLs[policyurl] = Result(embedurl);
	}

	for( pos=policyURLs; pos; ++pos )
	{
		CheckValid();
		TRY
		{
			g_policy_handler.HandlePolicyURL(policyURLs.key(pos), GURL(), policyURLs[pos]);
		}
		CATCH(ex)
		{
		}
		ENDCATCH;
	}
	}
	CATCH(ex){}ENDCATCH;
}


void P3PMatchInfo::Clean()
{
	CleanInfo();
	CleanCookies();
}

void P3PMatchInfo::CleanInfo()
{
	GCriticalSectionLock lock(&InfoLock);
	embededURLs.empty();
	policyURLs.empty();
	result = Result(GURL());	
}

void P3PMatchInfo::CleanCookies()
{
	GCriticalSectionLock lock(&cookielock);
	cookielist.empty();
}

P3PMatchInfo::~P3PMatchInfo()
{
	Clean();
}

P3PMatchInfo::P3PMatch P3PMatchInfo::GetNetResult(bool &blocked_cookies)
{
	CheckValid();
	{
		blocked_cookies = false;
		GCriticalSectionLock cooklock(&cookielock);
		for( GPosition pos = cookielist; pos; ++pos)
		{
			if( cookielist[pos]->blocked )
			{
				blocked_cookies = true;
				break;
			}
		}
	}


	GCriticalSectionLock lock(&InfoLock);
	P3PMatch match = UNKNOWN;
	if( result.Test )
		return match;

	bool anyunknowns = false;
	for(GPosition pos = embededURLs; pos; ++pos )
	{
		if( embededURLs[pos].URL.is_empty() )
		{
			anyunknowns = true;
			break;
		}
	}

	if( result.Match == Result::MATCH )
	{
		match = MATCH;
		if( anyunknowns )
		{
			match = MATCH_NOTHIRDPARTY;
		}
		else
		{
			for( GPosition pos=policyURLs; pos; ++pos )
			{
				if( policyURLs[pos].Match != Result::MATCH )
				{
					match = MATCH_NOTHIRDPARTY;
					break;
				}
			}
		}
	}
	else if( result.Match == Result::UNMATCH )
	{
		match = UNMATCH;
		/// nothing to do much

	}
	else // if unknown
	{
		match = UNKNOWN;
		/*
		match = UNKNOWN_MATCHTHIRDPARTY;
		if( !anyunknowns )
		{
			for( GPosition pos=policyURLs; pos; ++pos )
			{
				if( policyURLs[pos].Match != Result::MATCH )
				{
					match = UNKNOWN;
					break;
				}
			}
		}
		*/

	}
	return match;
}

void P3PMatchInfo::GetEmbeddedURLList(GList<GURL> &urllist)
{
	GCriticalSectionLock lock(&InfoLock);

	for( GPosition pos = embededURLs; pos; ++pos )
	{
		urllist.append( embededURLs.key(pos) );
	}
}

Result::P3PMatch P3PMatchInfo::GetEmbeddedResult(const GURL &url)
{
	GCriticalSectionLock lock(&InfoLock);
	if( GPosition pos = embededURLs.contains(url) )
	{
		if( pos = policyURLs.contains(embededURLs[pos].URL) )
		{
			return policyURLs[pos].Match;
		}
	}
	return Result::UNKNOWN;
}

GString P3PMatchInfo::GetEmbeddedURLType(const GURL &url)
{
	GCriticalSectionLock lock(&InfoLock);
	if( GPosition pos = embededURLs.contains(url) )
	{
		return embededURLs[pos].type;
	}
	return "";
}


bool P3PMatchInfo::GetEmbeddedResult(const GURL &embedURL, Result &result)
{
	GCriticalSectionLock lock(&InfoLock);
	if( GPosition pos = embededURLs.contains(embedURL) )
	{
		URLTypeInfo typeinfo = embededURLs[pos];
		if( typeinfo.URL.is_empty() )
			return false;

		if( pos = policyURLs.contains(typeinfo.URL) )
		{
			result = policyURLs[pos];
			return true;
		}
	}
	return false;
}


GURL P3PMatchInfo::GetPolicyURL(const GURL &embedURL)
{
	GCriticalSectionLock lock(&InfoLock);
	if( GPosition pos = embededURLs.contains(embedURL) )
	{
		URLTypeInfo typeinfo = embededURLs[pos];
		if( typeinfo.URL.is_empty() )
			return GURL();

		if( pos = policyURLs.contains(typeinfo.URL) )
		{
			return policyURLs.key(pos);
		}
	}
	return GURL();
}


int P3PMatchInfo::CheckMainURL(const GURL &url, GURL &policyURL)
{
	GCriticalSectionLock lock(&InfoLock);
	mainURL = url;
	result = Result(url);
	if( policyURL.is_empty() )
		policyURL = g_policy_handler.FindPolicy(url, "GET");
	if( !policyURL.is_empty() )
	{
		result.appliedPolicyURL = policyURL;
		return g_policy_handler.HandlePolicyURL(policyURL, url, result);
	}
	return -1;
}

GURL P3PMatchInfo::GetAppliedMainPolicyURL()
{
	GCriticalSectionLock lock(&InfoLock);
	return result.appliedPolicyURL;
}

int P3PMatchInfo::CheckMainURLWithP3PHeader(const GURL &siteurl)
{
	CheckValid();

	GString header;

	if( !GetP3PHeaderFromCache(siteurl, header) )
	{
		if( ((GString)siteurl).search('?') < 0 )
			GetPage(siteurl, header, true);
	}

	GURL refurl = GetURLFromP3PHeader(header, siteurl);
	if( !refurl.is_empty() )
	{
		g_policy_handler.HandlePolicyReferenceURL(refurl, siteurl);
		return CheckMainURL(siteurl);
	}
return -1;
}

////block cookie if not matched
bool P3PMatchInfo::AddCookie(const char *name, const char *value, const char *sitedomain, const char *trackerdomain, const char *path, const GURL &mainurl)
{
	GString domain;
	if( trackerdomain && strlen(trackerdomain))
		domain = trackerdomain;
	else
		domain = sitedomain;

	int retval = -1;
	GCriticalSectionLock lock(&cookielock);
	GP<CookieInfo> ci = new CookieInfo(name, value, domain, path);
	cookielist.append(ci);

	///add to history
	g_p3p_preferences.AddCookieHistory(sitedomain, trackerdomain);
	

	if( !mainurl.is_empty() )
	{
		if( patternmatch(GString("*.")+domain, mainurl.domain() ) )
		{
			/*
			if( GPosition pos = g_p3p_preferences.domain_prefs.contains(domain) )
			{
				P3PPreferences::site_allow_cookies cp = g_p3p_preferences.domain_prefs[pos];
				switch(cp)
				{
					case P3PPreferences::ALWAYS:
					case P3PPreferences::THIS_SESSION:
						ci->needpolicycheck = false;
						return false;

					case P3PPreferences::NEVER:
						ci->needpolicycheck = false;
						ci->blocked = true;
						return true;
				}
			}
			*/
		}
		else
		{
			ci->thirdparty = true;
		}
	}
	else
	{
		ci->thirdparty = true;
	}


	{
		if( GPosition pos = g_p3p_preferences.domain_prefs.contains(domain) )
		{
			P3PPreferences::site_allow_cookies cp = g_p3p_preferences.domain_prefs[pos];
			switch(cp)
			{
				case P3PPreferences::ALWAYS:
				case P3PPreferences::THIS_SESSION:
					ci->needpolicycheck = false;
					return false;

				case P3PPreferences::NEVER:
					ci->needpolicycheck = false;
					ci->blocked = true;
					return true;
			}
		}
	}


	{	
		switch(g_p3p_preferences.allow_cookies)
		{
		case P3PPreferences::MATCHING_PREFERENCES:
			{
				int retval = CheckCookiePolicy(name, value, domain,path);
				if( retval > 0 )
				{
					ci->blocked = true;
					ci->needpolicycheck = false;
				}
				else if( retval == 0 )
				{
					ci->blocked = false;
					ci->needpolicycheck = false;
				}
				else if( retval == -2 )
				{
					ci->blocked = true;
					ci->needpolicycheck = false;
				}
				else
				{
					ci->blocked = false;
					ci->needpolicycheck = true;
				}
			}
			return ci->blocked;

		case P3PPreferences::ALL:
			ci->needpolicycheck = false;
			return false;

		case P3PPreferences::NONE:
			ci->needpolicycheck = false;
			ci->blocked = true;
			return true;

		case P3PPreferences::CUSTOM_COOKIE:
			{
				int what=0;
				if( !mainurl.is_empty() && patternmatch(GString("*.")+domain, mainurl.domain() ) )
				{
					what = g_p3p_preferences.allow_cookies_custom;
				}
				else
				{
					what = g_p3p_preferences.allow_cookies_custom_crosssite;
				}

				switch(what)
				{
				case P3PPreferences::MATCHING_PREFERENCES:
					{
						int retval = CheckCookiePolicy(name, value, domain,path);
						if( retval > 0 )
						{
							ci->blocked = true;
							ci->needpolicycheck = false;
						}
						else if( retval == 0 )
						{
							ci->blocked = false;
							ci->needpolicycheck = false;
						}
						else if( retval == -2 )
						{
							ci->blocked = true;
							ci->needpolicycheck = false;
						}
						else
						{
							ci->blocked = false;
							ci->needpolicycheck = true;
						}
					}
					return ci->blocked;

				case P3PPreferences::ALL:
					ci->needpolicycheck = false;
					return false;

				case P3PPreferences::NONE:
					ci->needpolicycheck = false;
					ci->blocked = true;
					return true;
				}
			}
			break;
		}
	}	
	return ci->blocked;
}

int P3PMatchInfo::CheckCookiePolicy(const GString &name, const GString &value, const GString &domain, const GString &path, bool downloadifneeded)
{
	int retval = -1;

	GURL realurl = g_policy_handler.FindPolicyForCookie(name, value, domain, path);
	if( !realurl.is_empty() )
	{


		GURL polurl = realurl;
		GString policyname = SeparateNameAndURL(polurl);

		char *policydata=g_policy_handler.GetURL(polurl);
		
		///match policy to preferences
		if( policydata )
		{
			TRY
			{
				DArray<char> data;
				if( g_p3p_preferences.ReadPreferenceFile(data) )
				{
					

					Result result(domain);
					if( P3PEvaluator::P3PEvaluator_Match_Memory(data, data.size(), policydata, strlen(policydata), polurl, policyname, result) == A_OK )
					{
						retval = result.MisMatches.size();
					}
					else
					{
						retval = -2;
					}
				}
				else
					retval = -2;
			}
			CATCH(ex)
			{
				if( policydata )
				{
					delete policydata;
					policydata = 0;
				}
				RETHROW;
			}
			ENDCATCH;
		}
		else
			retval = -1;

		if( policydata )
			delete policydata;
	}

	return retval;
}

void P3PMatchInfo::CheckAllCookiePolicies()
{
	CheckValid();
	///create a copy to avoid locking lot of time
	GPList<CookieInfo> cookielist;
	{
		GCriticalSectionLock lock(&cookielock);
		cookielist = this->cookielist;
	}
	for( GPosition pos=cookielist; pos; ++pos)
	{
		CheckValid();

		GP<CookieInfo> ci = cookielist[pos];
		if( ci->needpolicycheck )
		{
			int retval = CheckCookiePolicy(ci->name, ci->value, ci->domain, ci->path, true);
			if( retval > 0 )
			{
				ci->blocked = true;
				ci->needpolicycheck = false;
			}
			else if( retval == 0 )
			{
				ci->blocked = false;
				ci->needpolicycheck = false;
			}
			else if( retval == -2 )
			{
				ci->blocked = true;
				ci->needpolicycheck = false;
			}
			else
			{
				ci->blocked = true;
				ci->needpolicycheck = false;
			}
		}
	}
}

	


bool P3PMatchInfo::NeedCookieBlocking(const char *name, const char *value, const char *sitedomain, const char *trackerdomain, const char *path, const GURL &mainurl)
{
	GString domain;
	if( trackerdomain && strlen(trackerdomain))
		domain = trackerdomain;
	else
		domain = sitedomain;

	//if( !mainurl.is_empty() )
	//{
	//	if( patternmatch(GString("*.")+domain, mainurl.domain() ) )
	//	{
			if( GPosition pos = g_p3p_preferences.domain_prefs.contains(domain) )
			{
				P3PPreferences::site_allow_cookies cp = g_p3p_preferences.domain_prefs[pos];
				switch(cp)
				{
					case P3PPreferences::ALWAYS:
					case P3PPreferences::THIS_SESSION:
						return false;

					case P3PPreferences::NEVER:
						return true;
				}
			}
	//	}
	//}
	{	
		switch(g_p3p_preferences.allow_cookies)
		{
		case P3PPreferences::MATCHING_PREFERENCES:
			if( CheckCookiePolicy(name, value, domain, path) )
				return true;
			else
				return false;

		case P3PPreferences::ALL:
			return false;

		case P3PPreferences::NONE:
			return true;

		case P3PPreferences::CUSTOM_COOKIE:
			{
				if( !mainurl.is_empty() && patternmatch(GString("*.")+domain, mainurl.domain() ) )
				{
					switch( g_p3p_preferences.allow_cookies_custom )
					{
						case P3PPreferences::MATCHING_PREFERENCES:
							if( CheckCookiePolicy(name, value, domain, path) )
								return true;
							else
								return false;

						case P3PPreferences::ALL:
							return false;

						case P3PPreferences::NONE:
							return true;
					}
				}
				else
				{
					///cross site cookies
					switch( g_p3p_preferences.allow_cookies_custom_crosssite )
					{
						case P3PPreferences::MATCHING_PREFERENCES:
							if( CheckCookiePolicy(name, value, domain, path) )
								return true;
							else
								return false;

						case P3PPreferences::ALL:
							return false;

						case P3PPreferences::NONE:
							return true;
					}
				}

			}
			break;
		}
	}
	return true;
}

void P3PMatchInfo::GetCookieInfo(int &cookiesdetected, int &cookiesblocked, int &othercookiesdetected, int &othercookiesblocked)
{
	
	cookiesdetected = cookiesblocked = othercookiesdetected = othercookiesblocked = 0;
	GCriticalSectionLock lock(&cookielock);
	for( GPosition pos=cookielist; pos; ++pos)
	{
		
		GP<CookieInfo> ci = cookielist[pos];
		if( ci->thirdparty )
		{
			othercookiesdetected++;
			if( ci->blocked )
				othercookiesblocked++;
		}
		else
		{
			cookiesdetected++;
			if( ci->blocked )
				cookiesblocked++;
		}
	}
}



///////////////////////////////////////////////////////////////////////
////static functions
///////////////////////////////////////////////////////////////////////
void P3PMatchInfo::PlaySoundPolicyUnkown()
{
	if(g_p3p_preferences.enable_sounds & P3PPreferences::ENABLE_SOUND_UNKNOWN)
		::PlaySound(MAKEINTRESOURCE(IDR_SOUND_UNKNOWN), hinst, SND_RESOURCE|SND_SYNC );
}

void P3PMatchInfo::PlaySoundPolicyUnmatched()
{
	if(g_p3p_preferences.enable_sounds & P3PPreferences::ENABLE_SOUND_UNMATCHED)
		::PlaySound(MAKEINTRESOURCE(IDR_SOUND_UNMATCH), hinst, SND_RESOURCE|SND_SYNC );
}

void P3PMatchInfo::PlaySoundPolicyMatched()
{
	if(g_p3p_preferences.enable_sounds & P3PPreferences::ENABLE_SOUND_MATCHED)
		::PlaySound(MAKEINTRESOURCE(IDR_SOUND_MATCH), hinst, SND_RESOURCE|SND_SYNC );
}

